/*
** Copyright Red Hat, Inc. 2013
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program; if not, write to the Free Software Foundation, Inc.,
** 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "Socket.h"
#include "String.h"
#include "Logger.h"  // questionable functionality split (+ op. separated here)

#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>


ServerSocket::ServerSocket(const String& sock_path) :
	Socket(-1),
	_unix_sock(true),
	_sock_path(sock_path)
{
	_sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (_sock == -1) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): socket() failed: " + String(strerror(errno));
	}

	int t = 1;
	if (setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, &t, sizeof(t))) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): set SO_REUSEADDR, failed: " + String(strerror(errno));
	}

	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;
	memcpy(addr.sun_path, sock_path.c_str(), sock_path.size() + 1);

	unlink(_sock_path.c_str());

	if (bind(_sock, (struct sockaddr*) &addr, sizeof(addr))) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): bind() failed: " + String(strerror(errno));
	}

	if (listen(_sock, 5)) {
		throw String("ServerSocket(sock_path=") + sock_path
				+ "): listen() failed: " + String(strerror(errno));
	}
	//String msg = String("created unix server socket, ")
	//				+ _sock + ", " + sock_path;
	//log(msg, LogSocket);
}

ServerSocket::ServerSocket(unsigned short port) :
	Socket(-1),
	_unix_sock(false),
	_sock_path("")
{
	/* Drop to pure IPv4 if IPv6 not applicable or cannot rely on dual
	   stack as indicated by getsockopt/setsockopt with IPV6_V6ONLY.
	   As a last resort, try pure IPv6 if still possible. */
	struct sockaddr_storage sa_addr;
	bool include_ip6 = true;
	_sock = socket(PF_INET6, SOCK_STREAM, 0);
	if (_sock != -1) {
		int i = -1;
		socklen_t il = sizeof(i);
		errno = 0;
		if (getsockopt(_sock, IPPROTO_IPV6, IPV6_V6ONLY, &i, &il) || i) {
			i = 0;
			if (errno
			  || setsockopt(_sock, IPPROTO_IPV6, IPV6_V6ONLY, &i, il)) {
				while (::close(_sock) && (errno == EINTR));
				_sock = -1;  // forget about IPv6 dual stack
			}  // otherwise we made dual stack ready for us
		}  // otherwise dual stack was already in place
	} else {
		include_ip6 = false;  // forget about IPv6
	}
	if (_sock == -1) {
		_sock = socket(PF_INET, SOCK_STREAM, 0);
		if (_sock != -1) {
			include_ip6 = false;  // IPv4 wins
		} else if (include_ip6) {
			_sock = socket(PF_INET6, SOCK_STREAM, 0);  // last resort
		}
		if (_sock == -1)
			throw String("ServerSocket(port=") + port
				+ "): socket() failed: " + String(strerror(errno));
	}

	int t = 1;
	if (setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, &t, sizeof(t))) {
		throw String("ServerSocket(port=") + port
			+ "): set SO_REUSEADDR, failed: " + String(strerror(errno));
	}

	if (!include_ip6) {
		struct sockaddr_in sin = {0};
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);
		sin.sin_addr.s_addr = INADDR_ANY;
		memcpy(&sa_addr, &sin, sizeof(sin));
	} else {
		struct sockaddr_in6 sin6 = {0};
		sin6.sin6_family = AF_INET6;
		sin6.sin6_port = htons(port);
		sin6.sin6_addr = in6addr_any;
		memcpy(&sa_addr, &sin6, sizeof(sin6));
	}

	if (bind(_sock, (const struct sockaddr *) &sa_addr, include_ip6
	         ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in))) {
		throw String("ServerSocket(port=") + port
			+ "): bind() failed: " + String(strerror(errno));
	}

	if (listen(_sock, 5)) {
		throw String("ServerSocket(port=") + port
			+ "): listen() failed: " + String(strerror(errno));
	}

	//String msg = String("created tcp server socket, ")
	//				+ _sock + ", port " + port;
	//log(msg, LogSocket);
}

ServerSocket::ServerSocket(const ServerSocket& s) :
	Socket(s),
	_unix_sock(s._unix_sock),
	_sock_path(s._sock_path)
{}

ServerSocket&
ServerSocket::operator= (const ServerSocket& s)
{
	if (&s != this) {
		this->Socket::operator= (s);
		_unix_sock = s._unix_sock;
		_sock_path = s._sock_path;
	}

	return *this;
}

ServerSocket::~ServerSocket()
{
	if (*_counter == 1) {
		if (_unix_sock)
			unlink(_sock_path.c_str());
	}
}

ClientSocket ServerSocket::accept() {
	while (true) {
		struct sockaddr_storage ss;
		socklen_t size = sizeof(ss);

		int ret = ::accept(_sock, (struct sockaddr *) &ss, &size);
		if (ret == -1) {
			if (errno == EINTR)
				continue;
			throw String("ServerSocket(): accept() failed: ")
					+ String(strerror(errno));
		}
		//log("ServerSocket: accepted connection", LogSocket);
		return ClientSocket(ret, (struct sockaddr *) &ss, size);
	}
}

bool
ServerSocket::ready(int timeout)
{
	bool read = true;
	bool write = false;

	poll(read, write, timeout);
	return read;
}
