/*
** Copyright 2016 Red Hat, Inc.
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#include "ClusterStatus.h"
#include "Socket.h"
#include "Time.h"
#include "ClusterConf.h"
#include "Clusvcadm.h"
#include "utils.h"

#include "String.h"
#include "ServiceManager.h"


using namespace std;

#define INITD_DIR_PATH			"/etc/init.d/"
#define LVMCONF_PATH			"/sbin/lvmconf"
#define UMOUNT_PATH				"/bin/umount"

#define CMAN_LEAVE_TIMEOUT		"120"	// seconds (string)
#define CLUMON_SYNC_TIME		8		// seconds
#define CMAN_SETTLE_TIME		3		// seconds


class Service;
class ServiceSet;
class ServiceManager;

static void run_initd(const String& servname);

bool
is_service_vm(const XMLObject& cluster_conf, const String& name)
{
	for (list<XMLObject>::const_iterator
			iter = cluster_conf.children().begin() ;
			iter != cluster_conf.children().end() ;
			iter++)
	{
		const XMLObject& kid = *iter;
		if (kid.tag() == "rm") {
			for (list<XMLObject>::const_iterator
					iter_s = kid.children().begin() ;
					iter_s != kid.children().end() ;
					iter_s++)
			{
				if (iter_s->tag() == "vm") {
					if (iter_s->get_attr("name") == name)
						return true;
				}
			}
		}
	}
	return false;
}

XMLObject
Cluster::status()
{
	ClientSocket sock;

	try {
		ClientSocket s("/var/run/clumond.sock");
		sock = s;
	} catch ( ... ) {
		// start clumon
		run_initd("modclusterd");

		// wait for it to come up and sync
		sleep_sec(CLUMON_SYNC_TIME);

		// try again
		ClientSocket s("/var/run/clumond.sock");
		sock = s;
	}
	sock.nonblocking(true);

	// send status request
	int beg = int(time_sec());
	String request("GET");

	while ((int(time_sec()) < beg + 5)) {
		bool read = false, write = true;
		sock.ready(read, write, 500);
		if (write) {
			if ((request = sock.send(request)).empty())
				break;
		}
	}

	// receive status report
	beg = int(time_sec());
	String xml_in;

	while ((int(time_sec()) < beg + 5)) {
		bool read = true, write = false;
		sock.ready(read, write, 500);
		if (read)
			xml_in += sock.recv();
		try {
			parseXML(xml_in);
			break;
		} catch ( ... ) {}
	}

	const XMLObject status_xml(parseXML(xml_in));
	const XMLObject cluster_conf(ClusterConf::get());

	if (cluster_conf.get_attr("name") != status_xml.get_attr("name"))
		throw String("cluster names mismatch");

	// add "vm" attr to services
	XMLObject status_new(status_xml.tag());
	for (map<String, String>::const_iterator
			iter = status_xml.attrs().begin() ;
			iter != status_xml.attrs().end() ;
			iter++)
	{
		status_new.set_attr(iter->first, iter->second);
	}

	for (list<XMLObject>::const_iterator
			iter = status_xml.children().begin() ;
			iter != status_xml.children().end() ;
			iter++)
	{
		XMLObject s(*iter);
		if (s.tag() == "service")
			s.set_attr("vm", (is_service_vm(cluster_conf,
								s.get_attr("name"))) ? "true" : "false");
		status_new.add_child(s);
	}
	return status_new;
}

#define CLUSTER6_SERVICES_LIST(_l)                         \
	do {                                               \
		(_l).push_back("cman");                    \
		(_l).push_back("cmirror");   /* RHEL 5 */  \
		(_l).push_back("cmirrord");  /* RHEL 6 */  \
		(_l).push_back("clvmd");                   \
		(_l).push_back("gfs2");                    \
		(_l).push_back("scsi_reserve");            \
		(_l).push_back("rgmanager");               \
	} while (0)


void
Cluster::start_node(bool cluster_startup, bool enable_services)
{
	// bail out if cluster.conf is not present
	XMLObject cluster_conf(ClusterConf::get());
	XMLObject stat = status();

	if (stat.get_attr("cluster_version") == "6") {
		list<String> svcs_str, svcs_setstr;
		list<Service> svcs_obj;
		list<ServiceSet> svcs_setobj;

		CLUSTER6_SERVICES_LIST(svcs_str);
		ServiceManager svcmgr = ServiceManager();
		svcmgr.query(svcs_str, svcs_setstr, svcs_obj, svcs_setobj);

		for (list<Service>::iterator
				iter = svcs_obj.begin() ;
				iter != svcs_obj.end() ;
				iter++)
		{
			try {
				iter->start();
			} catch ( ... ) {
				if (iter->name() == "cman")
					iter->start();  // try again
				else if (iter->name() == "rgmanager")
					throw;  // considered critical
			}
		}

		if (enable_services) {
			// enable them on boot
			for (list<Service>::iterator
					iter = svcs_obj.begin() ;
					iter != svcs_obj.end() ;
					iter++)
			{
				iter->enable();
			}
		}
	} else {
		throw String("unsupported cluster version ")
				+ stat.get_attr("cluster_version");
	}
}

void
Cluster::stop_node(	bool cluster_shutdown,
					bool purge_conf,
					bool disable_services)
{
	XMLObject stat = status();

	if (cluster_shutdown) {
		// stop all services, so they don't bounce around
		for (list<XMLObject>::const_iterator
			iter = stat.children().begin() ;
			iter != stat.children().end() ;
			iter++)
		{
			if (iter->tag() == "service") {
				if (iter->get_attr("running") == "true") {
					try {
						Clusvcadm::stop(iter->get_attr("name"));
					} catch ( ... ) {}
				}
			}
		}
	}

	if (stat.get_attr("cluster_version") == "6") {
		list<String> svcs_str, svcs_setstr;
		list<Service> svcs_obj;
		list<ServiceSet> svcs_setobj;
		vector<String> args;
		String out, err;
		int status;

		CLUSTER6_SERVICES_LIST(svcs_str);
		ServiceManager svcmgr = ServiceManager();
		svcmgr.query(svcs_str, svcs_setstr, svcs_obj, svcs_setobj);

		for (list<Service>::reverse_iterator
				iter(svcs_obj.end()) ;
				iter != list<Service>::reverse_iterator(svcs_obj.begin());
				iter++)
		{
			try {
				iter->stop();
			} catch ( ... ) {
				if (iter->name() == "cman"
				    || iter->name() == "rgmanager")
					throw;  // considered critical
			}
			if (iter->name() == "gfs2") {
				/*
				** Manually unmount any gfs2 filesystems not umounted by
				** either rgmanager of the gfs2 init script.
				*/
				args.push_back("-a");
				args.push_back("-t");
				args.push_back("gfs2");

				if (utils::execute(UMOUNT_PATH, args, out, err, status, false))
					throw command_not_found_error_msg(UMOUNT_PATH);

				if (status != 0)
					throw String("Error unmounting gfs2 filesystems: " + err);
			}
		}

		if (disable_services) {
			for (list<Service>::reverse_iterator
					iter(svcs_obj.end()) ;
					iter != list<Service>::reverse_iterator(svcs_obj.begin());
					iter++)
			{
				iter->disable();
			}
		}
	} else {
		throw String("unsupported cluster version ")
				+ stat.get_attr("cluster_version");
	}

	if (purge_conf) {
		ClusterConf::purge_conf();

		// disable LVM cluster locking
		try {
			String out, err;
			int status;
			vector<String> args;

			args.push_back("--disable-cluster");
			utils::execute(LVMCONF_PATH, args, out, err, status, false);
		} catch ( ... ) {}
	}
}

void
run_initd(const String& servname)
{
	String path(INITD_DIR_PATH);
	path += servname;

	String out, err;
	int status;
	vector<String> args;

	args.push_back("start");
	bool failed = true;

	if (utils::execute(path, args, out, err, status, false) == 0) {
		if (status == 0)
			failed = false;
	}

	if (failed) {
		throw String("service ") + servname + " "
				+ String("start failed: ") + err;
	}
}
